#include <cmath>
#include <zlib.h>
#include "kstring.h"//<-used for buffered output when dumping beagle 0.204
#include "bfgs.h"
#include "analysisFunction.h"
#include "general.h"
#include <vector>
#include "soap_likes.h"
#include "gatk_likes.h"
#include "bam_likes.h"

extern int refToInt[256];
std::vector<char*> getFilenames(const char * name,int nItems);
std::vector<char *> getBams(argStruct *arguments);

class likeClass:public general{
private:
  double **errors;
  char * errorFname;
  kstring_t bufstr;
  double ****errorProbs;

  soap_likes soap;
  int doingGLest;//if data needs recalibration, this is set to zero
  std::vector<char *> bamNames;
  //stuff for soapsnp model only
  int qcharmin; //minimum qs 
  int rlen; //the readlength, the legnth of the sequnce
  int MIN_BASE_q;//used for removing low qs in samtools GL estimattion
  int MIN_MAP_Q;//used for removing low mapQ in samtools GL estimattion
  char *angsd_tmpdir;
public:
  //none optional stuff
  FILE *outfile;
  gzFile gzoutfile;
  int GL;
  int doGlf;
  void run(funkyPars  *pars);
  void print(funkyPars *pars);  
  void clean(funkyPars *pars);  
  void addDefault(funkyPars *pars);
  void openfile(const char *outfiles);
  void getOptions(argStruct *arguments);
  void printArg(FILE *argFile,int method);

  int init(int nSamples);
  likeClass(const char *outfiles,argStruct *arguments,int inputtype);
  ~likeClass();
  //other stuff
  int pml_genLikes(funkyPars *pars);
  double **getLikesFullError10Genotypes(int numSites,int nInd,suint **counts,double ****errorProbs,int *keepSites);
  void printLike(funkyPars *pars);
};

/*
  readerors from an output files
  new version averages from all lines.
*/

void readError(double **errors,const char *fname){
  fprintf(stderr,"will try to read errorestimates from file:%s\n",fname);
  FILE *fp=NULL;
  if(NULL==(fp=fopen(fname,"r"))){
    fprintf(stderr,"Error opening file: %s\n",fname);
    exit(0);
  }
  
  char buf[LENS];
  double res[16];
  for(int i=0;i<16;i++) res[i] = 0;

  int nLines =0;
  while(fgets(buf,LENS,fp)){
    res[0] += atof(strtok(buf," \t\n"));
    for(int i=1;i<16;i++)
      res[i] += atof(strtok(NULL," \t\n"));
    nLines ++;
  }
  for(int j=0;j<16;j++)
    fprintf(stderr,"%f\t",res[j]);
  fprintf(stderr,"\nEstimating errors using nChunks:%d\n",nLines);
  int pos =0;
  for(int i=0;i<4;i++)
    for(int j=0;j<4;j++)
      errors[i][j] = res[pos++]/(1.0*nLines);
  fclose(fp);
}


void likeClass::printArg(FILE *argFile,int method){
  fprintf(argFile,"%s:\n\n",__FILE__);

  if(method==1||method==0){
    fprintf(argFile,"GL=1: samtools genotype likelihoods\n");
    fprintf(argFile,"\t-minQ\t%d\n",MIN_BASE_q);
    fprintf(argFile,"\t-minMapQ\t%d\n",MIN_MAP_Q);
    fprintf(argFile,"\n");
  }
  if(method==2||method==0){
    fprintf(argFile,"GL=2: gatk genotype likelihoods\n");
    fprintf(argFile,"\t-minQ\t%d\n",MIN_BASE_q);
    fprintf(argFile,"\t-minMapQ\t%d\n",MIN_MAP_Q);
    fprintf(argFile,"\n");
  }
  if(method==3||method==0){
    fprintf(argFile,"GL=3: soapSNP genotype likelihoods (MAX qs=255,max read_len=255)\n");
    fprintf(argFile,"\t-minQ\t%d\n",MIN_BASE_q);
    fprintf(argFile,"\t-minMapQ\t%d\n",MIN_MAP_Q);
    fprintf(argFile,"\t-tmpdir\t%s\n",angsd_tmpdir);
    fprintf(argFile,"\n");
  }
  if(method==4||method==0){
    fprintf(argFile,"GL=4: kim et. al. genotype likelihoods\n");
    fprintf(argFile,"\t-errors\t%s\n",errorFname);
    fprintf(argFile,"\n");
  }

  fprintf(argFile,"-doGlf\t%d\n",doGlf);
    fprintf(argFile,"\t1: binary glf (10 log likes)\n");
    fprintf(argFile,"\t2: beagle likelihood file\n");
    fprintf(argFile,"\t3: binary 3 times likelihood\n");
    fprintf(argFile,"\t4: text version (10 log likes)\n");
    fprintf(argFile,"\n");

}

void likeClass::getOptions(argStruct *arguments){

  
  //default
  GL=0;
  doGlf=0;
  errorFname = NULL;
  qcharmin=0;
  rlen=100;
  doingGLest =0;
  GL=0;//this means not to run
  MIN_BASE_q = 13;
  MIN_MAP_Q = 0;
  angsd_tmpdir = strdup("angsd_tmpdir");//leak, don't care

  int doMajorMinor=0;
  int doCounts =0;

  //from command line
  GL=angsd::getArg("-GL",GL,arguments);
  if(GL==-999){
    GL=0;
    printArg(stdout,0);
    exit(0);
  }
  if(GL==0)
    return;
  if(GL<-2||GL>4){
    fprintf(stderr,"You supplied: -GL %d, -1)mpileup  1)samtool 2)gatk 3)soap 4)kim\n",GL);
    exit(0);
  }
 
  if(GL>0&&GL<4){//all but kim
    MIN_BASE_q = angsd::getArg("-minQ",MIN_BASE_q,arguments);
    MIN_MAP_Q = angsd::getArg("-minMapQ",MIN_MAP_Q,arguments);
  }
  if(GL==3){//soap
    qcharmin=MIN_BASE_q;
    angsd_tmpdir = angsd::getArg("-tmpdir",angsd_tmpdir,arguments);
  }
  if(GL==4){// kim
    doCounts=angsd::getArg("-doCounts",doCounts,arguments);
    errorFname = angsd::getArg("-errors",errorFname,arguments);
  }
  doMajorMinor=angsd::getArg("-doMajorMinor",doMajorMinor,arguments);

  //all methods
  doGlf=angsd::getArg("-doGlf",doGlf,arguments);

  
  if(GL==4&&(doCounts==0)){
    fprintf(stderr,"Must supply -doCounts\n");
    exit(0);
  }
  if(doGlf>1 && doMajorMinor==0){
    fprintf(stderr,"For dumping beaglestyle output you need to estimate major/minor: -doMajorMinor\n");
    exit(0);
  }
  printArg(arguments->argumentFile,GL);
  bamNames = getBams(arguments);
}

likeClass::likeClass(const char *outfiles,argStruct *arguments,int inputtype){
  
  getOptions(arguments);
  if(GL==0)
    return;

  outfile = NULL;
  gzoutfile = Z_NULL;

  if(doGlf){
    const char* postfix = ".glf";
    const char *beaglepostfix = ".beagle.gz";
    if(doGlf!=2)
      outfile = openFile(outfiles,postfix);
    else{
      gzoutfile = openFileGz(outfiles,beaglepostfix,"w6h");
      bufstr.s=NULL; bufstr.l=bufstr.m=0;// <- used for buffered output 
      bufstr.l=0;
      kputs("marker\tallele1\tallele2",&bufstr);
      for(int i=0;i<arguments->nInd;i++){
	kputs("\tInd",&bufstr);
	kputw(i,&bufstr);
	kputs("\tInd",&bufstr);
	kputw(i,&bufstr);
	kputs("\tInd",&bufstr);
	kputw(i,&bufstr);
      }
      kputc('\n',&bufstr);
      gzwrite(gzoutfile,bufstr.s,bufstr.l);
    }
    // fprintf(stderr,"Dumping likelihoods in: %s\n",outfiles);
  }
 
  if(GL==4){
    errorProbs = NULL;

  //default errormatrix
    double errorsDefault[4][4]={    {0,       0.00031, 0.00373, 0.000664},
				  {0.000737    ,    0  ,     0.000576    ,    0.001702},
				  {0.001825 ,       0.000386  ,      0 ,      0.000653},
				  {0.00066, 0.003648 ,       0.000321  ,      0},
    };
    if(inputtype==5&&doGlf){
      fprintf(stderr,"Error: cannot output likelihoods (doGlf) when input is beagle\n");
      exit(0);
    }
    if(doGlf&&GL==0&&inputtype==1){
      fprintf(stderr,"You have do estimate likelihoods (-GL) for output (-doGlf) \n");
      exit(0);
    }
    //inputtype 0) soap 1) samglf 2) samglfClean 3) tglf 4)sim type 5) beagle 7) upppile
    if(inputtype!=0&&inputtype!=7){
      fprintf(stderr,"Error: Likelihoods can only be estimated based on SOAP input and uppile input\n");
      exit(0);
    }
    //allocate and plug in default values
    errors = new double *[4];
    for(int i=0;i<4;i++){
      errors[i] = new double[4];
      for(int j=0;j<4;j++)
	errors[i][j] = errorsDefault[i][j];
    }
    if(errorFname!=NULL)
      readError(errors,errorFname);
    errorProbs = error::generateErrorPointers(errors,3,4);
  }
  if(GL==3){
    doingGLest= soap.init(bamNames,angsd_tmpdir);
    if(doingGLest)
      fprintf(stderr,"[%s] Will use precalculated calibration matrices\n",__FILE__);
    else
      fprintf(stderr,"[%s] Will calculate recalibration matrices, please don't do any other analysis\n",__FILE__);

  }
  if(GL==1)
    bam_likes_init();
   //   fprintf(stderr,"perform GL estimation=%d\n",doingGLest); 

}


likeClass::~likeClass(){
  if(GL==4)
    error::killGlobalErrorProbs(errorProbs);
  if(doGlf)//filehandle is only open if we want to print GL
    if(outfile!=NULL)
      fclose(outfile);
    else
      gzclose(gzoutfile);
  
  //  run(NULL);
  for(uint i=0;i<bamNames.size();i++)
    free(bamNames[i]);

}



void likeClass::clean(funkyPars *pars){

  if(pars->likes!=NULL){
    for(int i=0;i<pars->numSites;i++)
      delete [] pars->likes[i];
    delete [] pars->likes;
  }
  if(pars->post!=NULL){
    for(int i=0;i<pars->numSites;i++)
      delete [] pars->post[i];
    delete [] pars->post;
  }
  if(GL!=4)
    return;
}


void likeClass::print(funkyPars *pars){

  if(doGlf){
    printLike(pars);
  }
  
}


void likeClass::run(funkyPars *pars){
  if(pars==NULL){
    fprintf(stderr,"NULL funkypars received in likeclass::run\n");
    return;
  }  
  if(GL==0)
    return;
  if(GL==4){
    pml_genLikes(pars);
    return;
  }
  
  chunkyT *chk;
  char *refs;
  chk=pars->chk;
  refs=pars->ref;
  if(chk==NULL)
    return;

  double **likes = NULL;
  likes = new double*[chk->nSites];
   
  if(GL==3){
    soap.run(chk,likes,refs,MIN_BASE_q,MIN_MAP_Q);
    if(doingGLest==0){//we dont estimate GL but make a calibration matrix
      delete [] likes;
      return;
    }
  }else if(GL==1)
    call_bam(chk,likes,MIN_BASE_q,MIN_MAP_Q);
  else if(GL==2)
    call_gatk(chk,likes,MIN_MAP_Q,MIN_BASE_q);
  
  pars->likes = likes;


}


int likeClass::pml_genLikes(funkyPars *pars){

  pars->likes = getLikesFullError10Genotypes(pars->numSites,pars->nInd,pars->counts,errorProbs,pars->keepSites);
  return 0;
}

double **likeClass::getLikesFullError10Genotypes(int numSites,int nInd,suint **counts,double ****errorProbs,int *keepSites) {

  //only calculate this once
  static float *logfactorial=NULL;
  if(logfactorial==NULL)//dont bother populating if exists.
    logfactorial=error::logfact(LOGMAX); //calculate log factorials

  double **loglikes = new double*[numSites];
  double *logError;


  for(int s=0;s<numSites;s++){
    loglikes[s] = new double [10*nInd];
    if(keepSites[s]==0)
      continue;
    for(int allele1=0;allele1<4;allele1++) {
      for(int allele2=allele1;allele2<4;allele2++){
	int Gindex=angsd::majorminor[allele1][allele2];
	int geno=0;
	int m2=allele2;
	if(allele1!=allele2)
	  geno++;
	else{//total grimt must redo
	m2++;
	 if(m2>3)
	  m2=0;
	}
	logError=errorProbs[geno][allele1][m2];
	for(int i=0;i<nInd;i++){
	  loglikes[s][i*10+Gindex]=logfactorial[counts[s][i*4+0]+counts[s][i*4+1]+counts[s][i*4+2]+counts[s][i*4+3]]; //should be computed before these loops for faster implimentation
	  for(int j=0;j<4;j++)
	    loglikes[s][i*10+Gindex]+=-logfactorial[counts[s][i*4+j]]+counts[s][i*4+j]*logError[j];
	}
      }
    }




    if(0&&s==2){//print infomation about counts and computed likelihood
      for(int i=0;i<nInd;i++){
	fprintf(stderr,"%d\t%d\t%d\t%d\t | %f\t%f\t%f\t",counts[s][i*4+0],counts[s][i*4+1],counts[s][i*4+2],counts[s][i*4+3],loglikes[s][i*3],loglikes[s][i*3+1],loglikes[s][i*3+2]);
	for(int j=0;j<10;j++)
	  fprintf(stderr,"%f\t",loglikes[s][i*10+j]);
	  fprintf(stderr,"\n");
      }


      exit(0);
    }
  }
  return loglikes;
}



void likeClass::printLike(funkyPars *pars) {
  if(pars->likes==NULL)
    return;
  //  pars->opt->printLike=2;
  //fprintf(stderr,"printing likes with option %d\n",pars->opt->doGLF);

  //glffinn format
  if(doGlf==1){
    for(int i=0;i<pars->numSites;i++){
      if(pars->keepSites[i]==0)
	continue;
      fwrite(pars->likes[i],sizeof(double),10*pars->nInd,outfile);
    }
    fflush(outfile);
    //    exit(0);
  }
   else if(doGlf==2){  //beagle format
    
    //header of the beagle file

      //      fprintf(stderr,"outfile=%p\n",outfile);
   
 
   
      for(int s=0;s<pars->numSites;s++) {
	bufstr.l = 0; //set tmpbuf beginning to zero
	if(pars->keepSites[s]==0)
	  continue;
	kputs(pars->sites[s].chromo,&bufstr);
	kputc('_',&bufstr);
	kputw(pars->sites[s].position+1,&bufstr);
	kputc('\t',&bufstr);
	kputw(pars->major[s],&bufstr);
	kputc('\t',&bufstr);
	kputw(pars->minor[s],&bufstr);
	//	fprintf(outfile,"%s_%d\t%d\t%d",pars->sites[s].chromo,pars->sites[s].position,pars->major[s],pars->minor[s]);
	for(int i=0;i<pars->nInd;i++) {
	  double norm=exp(pars->likes[s][i*10+angsd::majorminor[pars->major[s]][pars->major[s]]])+exp(pars->likes[s][i*10+angsd::majorminor[pars->major[s]][pars->minor[s]]])+exp(pars->likes[s][i*10+angsd::majorminor[pars->minor[s]][pars->minor[s]]]);
	  double val1 = exp(pars->likes[s][i*10+angsd::majorminor[pars->major[s]][pars->major[s]]])/norm;
	  double val2 = exp(pars->likes[s][i*10+angsd::majorminor[pars->major[s]][pars->minor[s]]])/norm;
	  double val3 = exp(pars->likes[s][i*10+angsd::majorminor[pars->minor[s]][pars->minor[s]]])/norm;
	  ksprintf(&bufstr, "\t%f",val1);
	  ksprintf(&bufstr, "\t%f",val2);
	  ksprintf(&bufstr, "\t%f",val3);
	  //fprintf(outfile,"\t%f",exp(pars->likes[s][i*10+angsd::majorminor[pars->major[s]][pars->major[s]]])/norm);
	  //fprintf(outfile,"\t%f",exp(pars->likes[s][i*10+angsd::majorminor[pars->major[s]][pars->minor[s]]])/norm);
	  //fprintf(outfile,"\t%f",exp(pars->likes[s][i*10+angsd::majorminor[pars->minor[s]][pars->minor[s]]])/norm);
	}
	//loglike[s][i]);
	kputc('\n',&bufstr);
	if(outfile!=NULL)
	  fprintf(outfile,"%s",bufstr.s);
	else
	  gzwrite(gzoutfile,bufstr.s,bufstr.l);
      }
    

    ////////
  }  else if(doGlf==3) { //FGV v0.208 Aug,28
    for(int s=0;s<pars->numSites;s++) {
      if(pars->keepSites[s]==0) //TSK 0.441 sep 25
	continue;
      for(int i=0;i<pars->nInd;i++) {
	fwrite(&pars->likes[s][i*10+angsd::majorminor[pars->major[s]][pars->major[s]]], sizeof(double),1,outfile);
	fwrite(&pars->likes[s][i*10+angsd::majorminor[pars->major[s]][pars->minor[s]]], sizeof(double),1,outfile);
	fwrite(&pars->likes[s][i*10+angsd::majorminor[pars->minor[s]][pars->minor[s]]], sizeof(double),1,outfile);
      }
    }
  }



  else{
    //otherwise print textoutput
    //    fprintf(stderr,"[%s]length=%lu\n",__FUNCTION__,10*pars->nInd);
    for(int s=0;s<pars->numSites;s++){
      if(pars->keepSites[s]==0)
	continue;
      fprintf(outfile,"%s\t%d\t",pars->sites[s].chromo,pars->sites[s].position+1);
      for(int i=0;i<10*pars->nInd-1;i++)
	fprintf(outfile,"%f\t",pars->likes[s][i]);
      fprintf(outfile,"%f\n",pars->likes[s][10*pars->nInd-1]);
    }
  }
  fflush(outfile);
}


std::vector<char *> getBams(argStruct *arguments){

  std::vector<char *> ret;
  //first find a .bam file
  for(int i=0;i<arguments->argc;i++)
    if(strstr(arguments->argv[i],".bam"))
      ret.push_back(strdup(arguments->argv[i]));
  int singleFile = ret.size();
  //  fprintf(stderr,"singlife=%d\n",singleFile);
  const char *fff = NULL;
  int nInd =-1;
  for(int i=0;i<arguments->argc;i++) {
    if(!strcmp(arguments->argv[i],"-bam")){
      if(singleFile){
	fprintf(stderr,"Youve supplied  a .bam file, and also -b will exit\n");
	exit(0);
      }
      if((i+1)<arguments->argc)
	fff = arguments->argv[i+1];
      else{
	fprintf(stderr,"Problems -b supplied as last argumens will exti\n");
	exit(0);
      }
    }
    if(!strcmp(arguments->argv[i],"-nInd")){
      if((i+1)<arguments->argc)
	nInd = atoi(arguments->argv[i+1]);
      else{
	fprintf(stderr,"Problems -nInd supplied as last argumens will exti\n");
	exit(0);
      }
    }
  }
  if((fff!=NULL)&&!singleFile)
    ret = getFilenames(fff,nInd);
  return ret;
}
