#include "shared.h"
#include "beagleReader.h"
#include <sys/stat.h>
#include "zlib.h"
#include "shared.h"
#include "analysisFunction.h"

void beagle_reader::printArg(FILE *argFile){
  fprintf(argFile,"%s:\n\n",__FILE__);
  fprintf(argFile,"-beagle=%s \n",fname);
  fprintf(argFile,"\n");
}
void beagle_reader::getOptions(argStruct *arguments){
  fname=angsd::getArg("-fname",fname,arguments);
  if(strcmp(fname,"-999")==0){
    printArg(stdout);
    exit(0);
  }
  printArg(stdout);
}





void beagle_reader::init(argStruct *arguments){
  getOptions(arguments);
  intName = 1;// intrepret SNP name as chr_pos


  if(!fexists(fname)) {
    fprintf(stderr,"file %s does not exits\n",fname);
    exit(0);
  }
  openBeagleFile = gzopen(fname,"rb");
  //  fprintf(stderr,"beagle file is open\n");
  //fflush(stderr);
  int lens = 100000;
  char buffer[lens]; 
  const char *delims = "\t \n";

  int nCol=1;

  gzgets(openBeagleFile,buffer,lens); 
  strtok(buffer,delims);
  while(strtok(NULL,delims))
    nCol++;
  if(nCol % 3 ){
    fprintf(stderr,"\t-> Number of columns should be a multiple of 3, nCol=%d\n",nCol);
    exit(0);
  } 
  nInd=nCol/3-1;
  arguments->nInd=nInd;
  //fprintf(stderr,"nCol=%d\tnSamples=%d\n",nCol,nInd);

}

funkyPars *beagle_reader::fetch(int nIndOther,int chunksize){

int refToInt[256] = {
  0,1,2,3,4,4,4,4,4,4,4,4,4,4,4,4,//15
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//31
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//47
  0,1,2,3,4,4,4,4,4,4,4,4,4,4,4,4,//63
  4,0,4,1,4,4,4,2,4,4,4,4,4,4,4,4,//79
  4,4,4,4,3,4,4,4,4,4,4,4,4,4,4,4,//95
  4,0,4,1,4,4,4,2,4,4,4,4,4,4,4,4,//111
  4,4,4,4,3,4,4,4,4,4,4,4,4,4,4,4,//127
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//143
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//159
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//175
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//191
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//207
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//223
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,//239
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4//255
};

 int lens = 100000;
 char buffer[lens]; 
 const char *delims = "\t \n";
 const char *delims2 = "_\t \n";
 double **post = new double*[chunksize];

 funkyPars * myfunky = allocFunkyPars();
 myfunky->sites = new loci[chunksize];
 myfunky->major = new int[chunksize];
 myfunky->minor = new int[chunksize];

 
 for(int s=0;s<chunksize;s++)
   post[s] = new double[nInd*3];

 int nSites=0;
 intName=1;
 while(gzgets(openBeagleFile,buffer,lens)){
   
   if(intName){
     myfunky->sites[nSites].chromo = strdup(strtok(buffer,delims2));
     myfunky->sites[nSites].position = atoi(strtok(NULL,delims2));
   }
   else{
     myfunky->sites[nSites].chromo = strdup(strtok(buffer,delims));
     myfunky->sites[nSites].position = 0;
   }
  
   // char *SNPname=strtok(buffer,delims);


   myfunky->major[nSites] = refToInt[strtok(NULL,delims)[0]];
   myfunky->minor[nSites] = refToInt[strtok(NULL,delims)[0]];
   for(int i=0;i<nInd*3;i++)
     post[nSites][i] = atof(strtok(NULL,delims));

   nSites++;
   if(nSites>=chunksize)
     break;
 }

 if(nSites<chunksize){
   for(int s=nSites;s<chunksize;s++)
     delete[] post[s];
 }
 myfunky->nInd=nInd;
 myfunky->post=post;
 myfunky->numSites = nSites;
 if(nSites==0)
   return(NULL);
 return(myfunky);
}

int beagle_reader::fexists(const char* str){///@param str Filename given as a string.
  struct stat buffer ;
  return (stat(str, &buffer )==0 ); /// @return Function returns 1 if file exists.
}
